@@ -242,7 +242,7 @@ def deserialize_tree(
 ) -> PyTree:
   """Deserializes a PyTree to the same structure as `target`."""
 
-  def fn(keypath, _):
+  def _reconstruct_from_keypath(keypath, _):
     result = serialized
     for key in keypath:
       key_name = get_key_name(key)
@@ -253,7 +253,9 @@ def deserialize_tree(
     return result
 
   return jax.tree_util.tree_map_with_path(
-      fn, target, is_leaf=is_empty_or_leaf if keep_empty_nodes else None
+      _reconstruct_from_keypath,
+      target,
+      is_leaf=is_empty_or_leaf if keep_empty_nodes else None,
   )
 
 
@@ -305,12 +307,17 @@ def name_from_leaf_placeholder(placeholder: str) -> str:
     raise ValueError('Found placeholder beginning with unexpected prefix.')
 
 
+def is_supported_empty_aggregation_type(value: Any) -> bool:
+  """Determines if the *empty* value is supported for aggregation."""
+  return isinstance(value, (dict, list, type(None))) and not value
+
+
 def is_supported_aggregation_type(value: Any) -> bool:
   """Determines if the value is supported for aggregation."""
   return isinstance(
       value,
-      (str, int, float, np.number, np.ndarray, jnp.ndarray, bytes, type(None)),
-  ) or (isinstance(value, (dict, list)) and not value)
+      (str, int, float, np.number, np.ndarray, jnp.ndarray, bytes),
+  ) or is_supported_empty_aggregation_type(value)
 
 
 def pytree_structure(directory: epath.PathLike) -> PyTree:
@@ -516,13 +523,26 @@ def step_from_checkpoint_name(name: str) -> int:
   raise ValueError(f'Unrecognized name format: {name}.')
 
 
+def checkpoint_steps_paths(
+    checkpoint_dir: epath.PathLike,
+) -> List[epath.Path]:
+  """Returns a list of finalized checkpoint paths in the directory."""
+  checkpoint_dir = epath.Path(checkpoint_dir)
+  if not checkpoint_dir.exists():
+    raise ValueError(f'Path {checkpoint_dir} does not exist.')
+  return [
+      step_dir
+      for step_dir in checkpoint_dir.iterdir()
+      if _is_step_checkpoint(step_dir) and is_checkpoint_finalized(step_dir)
+  ]
+
+
 def checkpoint_steps(checkpoint_dir: epath.PathLike) -> List[int]:
   """Returns a list of finalized checkpoint steps in the directory."""
   checkpoint_dir = epath.Path(checkpoint_dir)
   return [
       step_from_checkpoint_name(s.name)
-      for s in checkpoint_dir.iterdir()
-      if _is_step_checkpoint(s) and is_checkpoint_finalized(s)
+      for s in checkpoint_steps_paths(checkpoint_dir)
   ]
 
 
